var // const
CrashRecoveryService = {
	mCID: Components.ID("{1280606b-2510-4fe0-97ef-9b5a22eafe6b}"),
	mContractID: "@zeniko/crashrecoveryservice;1",
	mClassName: "crashrecoveryService",

	mObserverService: Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService),
	mPrefService: Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService),
	mWindowMediator: Components.classes["@mozilla.org/appshell/window-mediator;1"].getService(Components.interfaces.nsIWindowMediator),
	mIOService: Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService),
	mPromptService: Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService),
	mDirectoryService: Components.classes["@mozilla.org/file/directory_service;1"].getService(Components.interfaces.nsIProperties),
	mUnicodeConverter: Components.classes["@mozilla.org/intl/scriptableunicodeconverter"].createInstance(Components.interfaces.nsIScriptableUnicodeConverter),

	mSaveInterval: 10000,
	mTabSeparator: "\n-- TAB ----\n",
	mLoadState: 0, // 1 = ready, -1 = shutting down
	mLastWindowClosed: "",
	mLastSaveTime: 0,
	mSavePending: false,
	mWindows: {},
	mCookies: [],

/* ........ Global Event Handlers .............. */

	init: function()
	{
		this.mObserverService.addObserver(this, "domwindowopened", false);
		this.mObserverService.addObserver(this, "domwindowclosed", false);
		this.mObserverService.addObserver(this, "quit-application-requested", false);
		this.mObserverService.addObserver(this, "quit-application-granted", false);
		this.mObserverService.addObserver(this, "quit-application", false);
		
		this.mPrefBranch = this.mPrefService.getBranch("");
		this.mStringBundle = getStringBundle("chrome://crashrecovery/locale/crashrecovery.properties");
		this.mUnicodeConverter.charset = "UTF-8";
		
		this.mInitialState = this.stateFromDisk();
		this.mLastSessionCrashed = /^.*~{4} /.test(this.mInitialState);
		this.mSaveInterval = this.getPref(".interval", 10000);
		if (typeof(this.mSaveInterval) != "number")
		{
			this.mSaveInterval = 10000;
		}
		if (this.mLastSessionCrashed)
		{
			try
			{
				this.restoreCache();
			}
			catch (ex) { }
		}
	},

	observe: function(aSubject, aTopic, aData)
	{
		switch (aTopic)
		{
		case "app-startup":
			this.mObserverService.addObserver(this, "profile-after-change", false);
			break;
		case "profile-after-change":
			this.init();
			break;
		case "domwindowopened":
			var window = aSubject;
			window.addEventListener("load", function() {
				CrashRecoveryService.onLoad(window);
			}, false);
			break;
		case "domwindowclosed":
			this.onClose(aSubject);
			break;
		case "quit-application-requested":
			this.getWindows("navigator:browser").forEach(function(aWindow) {
				CrashRecoveryService.saveWindowURLs(aWindow);
				CrashRecoveryService.updateTextAndFramesetData();
			});
			break;
		case "quit-application":
			if (this.mLoadState == -1)
			{
				break;
			}
		case "quit-application-granted":
			this.mLoadState = -1;
			if (this.doResumeSession())
			{
				this.saveState(true);
			}
			else
			{
				this.clearDisk();
			}
			break;
		}
	},

/* ........ Window Event Handlers .............. */

	onLoad: function(aWindow)
	{
		if (aWindow.document.documentElement.getAttribute("windowtype") == "navigator:browser")
		{
			var ix = this.getNextIndex()
			aWindow.__CRi = ix;
			this.mWindows[ix] = { tabs: [], index: -1 };
			
			if (!this.mLoadState && ((this.mLastSessionCrashed)?this.doRecoverSession():this.doResumeSession()))
			{
				this.restoreWindow(aWindow, this.mInitialState, this.isCmdLineEmpty(aWindow));
			}
			else
			{
				this.onLoad_done(aWindow);
			}
		}
	},

	onLoad_done: function(aWindow)
	{
		if (!this.mLoadState)
		{
			aWindow.setTimeout(function(aCRS) { aCRS.retryDownloads(); }, 100, this);
			if (this.mInitialState)
			{
				this.stateToDisk(this.mInitialState, true);
			}
			delete this.mInitialState;
			
			this.mLastSaveTime = (new Date()).valueOf() - this.mSaveInterval + 3000;
			this.mLoadState = 1;
			this.saveState();
		}
		
		aWindow.getBrowser().addEventListener("load", function(aEvent) { CrashRecoveryService.onTabLoad(aWindow, aEvent); }, true);
		aWindow.getBrowser().addEventListener("select", function(aEvent) { CrashRecoveryService.onTabSelect(aWindow, aEvent); }, false);
		aWindow.getBrowser().addEventListener("input", function(aEvent) { CrashRecoveryService.onTabInput(aWindow, aEvent); }, true);
		aWindow.getBrowser().addEventListener("DOMAutoComplete", function(aEvent) { CrashRecoveryService.onTabInput(aWindow, aEvent); }, true);
		aWindow.getBrowser().addEventListener("unload", function(aEvent) { CrashRecoveryService.onTabUnload(aWindow, aEvent); }, true);
	},

	onClose: function(aWindow)
	{
		if (aWindow.document.documentElement.getAttribute("windowtype") == "navigator:browser")
		{
			this.mLastWindowClosed = this.mWindows[aWindow.__CRi];
			delete this.mWindows[aWindow.__CRi];
		}
	},

	onTabLoad: function(aWindow, aEvent, aCount)
	{
		if (aEvent && aEvent.originalTarget.defaultView && aEvent.originalTarget.defaultView != aEvent.originalTarget.defaultView.top)
		{
			return;
		}
		
		if (aWindow.content == null)
		{
			aCount = (aCount || 0) + 1;
			if (aCount < 10)
			{
				aWindow.setTimeout(function(aCRS) { aCRS.onTabLoad(aWindow, null, aCount); }, 100, this);
			}
			return;
		}
		
		this.saveWindowURLs(aWindow);
	},

	onTabSelect: function(aWindow, aEvent)
	{
		this.saveWindowURLs(aWindow);
	},

	onTabInput: function(aWindow, aEvent)
	{
		this.saveTextData(aWindow, aEvent.originalTarget);
	},

	onTabUnload: function(aWindow, aEvent)
	{
		if (aEvent.target == aEvent.originalTarget)
		{
			this.saveWindowURLs(aWindow);
		}
	},

/* ........ Saving Functionality .............. */

	saveWindowURLs: function(aWindow)
	{
		if (this.mLoadState < 1 || !aWindow.__CRi)
		{
			return;
		}
		
		var browser = aWindow.getBrowser();
		var tabs = this.mWindows[aWindow.__CRi].tabs = [];
		this.mWindows[aWindow.__CRi].index = 0;
		
		for (var i = 0; i < browser.browsers.length; i++)
		{
			var tab = browser.browsers[i];
			var hist = null;
			var tabData = [];
			
			try
			{
				hist = tab.sessionHistory;
			}
			catch (ex)
			{
				continue;
			}
			
			if (hist && hist.count > 0)
			{
				tabData.push(hist.index + " zoom:" + tab.markupDocumentViewer.textZoom);
				
				for (var j = 0; j < hist.count; j++)
				{
					var entry = hist.getEntryAtIndex(j, false).QueryInterface(Components.interfaces.nsISHEntry);
					var info = [entry.URI.spec.replace(" ", "%20")];
					
					if (j == hist.index)
					{
						info.push(tab.contentWindow.scrollX + "," + tab.contentWindow.scrollY);
					}
					else
					{
						var x = {}, y = {};
						entry.getScrollPosition(x, y);
						info.push(x.value + "," + y.value);
					}
					
					if (entry.postData)
					{
						try
						{
							entry.postData.QueryInterface(Components.interfaces.nsISeekableStream).seek(Components.interfaces.nsISeekableStream.NS_SEEK_SET, 0);
							var stream = Components.classes["@mozilla.org/scriptableinputstream;1"].createInstance(Components.interfaces.nsIScriptableInputStream);
							stream.init(entry.postData);
							info.push(encodeURI(stream.read(stream.available())));
							stream.close();
						}
						catch (ex) { }
					}
					
					tabData.push(info.join(" "));
				}
			}
			else
			{
				tabData.push("0 zoom:1");
				
				info = [];
				info.push(tab.contentDocument.location.href.replace(" ", "%20"));
				info.push("0,0");
				
				tabData.push(info.join(" "));
			}
			
			tabs.push(tabData);
			
			if (tab == browser.mCurrentBrowser)
			{
				this.mWindows[aWindow.__CRi].index = i;
			}
		}
		
		this.saveStateDelayed(aWindow);
	},

	saveTextData: function(aWindow, aTextarea)
	{
		var id = (aTextarea.id)?"#" + aTextarea.id:aTextarea.name;
		if (!(id && (/^textarea$/i.test(aTextarea.tagName) || (/^input$/i.test(aTextarea.tagName) && /^text$/i.test(aTextarea.type)))))
		{
			return;
		}
		
		var document = aTextarea.ownerDocument.defaultView.top.document;
		if (!document.__CR_data)
		{
			document.__CR_data = {};
		}
		document.__CR_data[id] = aTextarea.value;
		
		this.saveStateDelayed(aWindow, 3000);
	},

	updateTextAndFramesetData: function()
	{
		var windows = {};
		
		this.forEachDocument(function(aWindow, aDocument) {
			if (!windows[aWindow.__CRi])
			{
				windows[aWindow.__CRi] = 0;
			}
			var tabData = CrashRecoveryService.mWindows[aWindow.__CRi].tabs[windows[aWindow.__CRi]] || null;
			if (!tabData)
			{
				windows[aWindow.__CRi]++;
				return;
			}
			
			var text = "";
			if (aDocument.__CR_data)
			{
				for (var id in aDocument.__CR_data)
				{
					text += " " + id + "=" + encodeURI(aDocument.__CR_data[id]);
				}
			}
			tabData.text = text.substr(1);
			
			var frameset = "";
			for (var i = 0; i < aDocument.defaultView.frames.length; i++)
			{
				frameset += " " + aDocument.defaultView.frames[i].location.href;
			}
			tabData.frameset = frameset.substr(1);
			
			windows[aWindow.__CRi]++;
		});
	},

	getCookieHosts: function()
	{
		var hosts = [];
		
		function extractHosts(aHost)
		{
			while (/^(?:[\w-]+\.([\w.-]+)|localhost)$/.test(aHost))
			{
				hosts.push(aHost);
				aHost = RegExp.$1;
			}
		}
		
		if (this.getNextIndex() == 1)
		{
			this.mWindows[-1] = this.mLastWindowClosed;
		}
		for (var ix in this.mWindows)
		{
			this.mWindows[ix].tabs.forEach(function(aTabData) {
				aTabData.slice(1).forEach(function(aData) {
					if (/^@frameset (.*)/.test(aData))
					{
						RegExp.$1.split(" ").forEach(function(aURL) {
							if (/^https?:\/\/(?:[^@\/\s]+@)?([\w.-]+)/.test(aURL))
							{
								extractHosts(RegExp.$1);
							}
						});
					}
					else if (/^(?!@text )https?:\/\/(?:[^@\/\s]+@)?([\w.-]+)\S* \d+,\d+/.test(aData))
					{
						extractHosts(RegExp.$1);
					}
				});
			});
		}
		if (this.mWindows[-1])
		{
			delete this.mWindows[-1];
		}
		
		return hosts;
	},

	saveCookies: function()
	{
		var cookies = Components.classes["@mozilla.org/cookiemanager;1"].getService(Components.interfaces.nsICookieManager).enumerator;
		var hosts = this.getCookieHosts()
		var sessionCookies = this.mCookies = [];
		
		while (cookies.hasMoreElements())
		{
			var cookie = cookies.getNext().QueryInterface(Components.interfaces.nsICookie);
			if (cookie.expires == 0 && cookie.host)
			{
				var host = cookie.host.replace(/^\./, "");
				if (hosts.indexOf(host) > -1)
				{
					var value = (cookie.name || "name") + "=" + (cookie.value || "") + ";";
					var domain = (cookie.isDomain)?"domain=" + host + ";":"";
					var path = (cookie.path)?"path=" + cookie.path + ";":"";
					var secure = (cookie.isSecure)?"secure;":"";
					var url = "http" + ((cookie.isSecure)?"s":"") + "://" + cookie.host + (cookie.path || "").replace(/^(?!\/)/, "/");
					sessionCookies.push(url + " " + encodeURI(value + domain + path + secure));
				}
			}
		}
	},

	getCurrentState: function()
	{
		if (this.mLoadState == 1)
		{
			this.updateTextAndFramesetData();
		}
		this.saveCookies();
		
		var total = [], windows = [];
		var activeTab = 0;
		for (var ix in this.mWindows)
		{
			if (this.mWindows[ix])
			{
				total.push(this.tabsToString(this.mWindows[ix].tabs));
				windows.push(parseInt(ix));
			}
		}
		
		var window = this.mWindowMediator.getMostRecentWindow("navigator:browser");
		if (window && windows.indexOf(window.__CRi) > -1)
		{
			var ix = windows.indexOf(window.__CRi);
			if (ix >= 0)
			{
				activeTab = this.mWindows[window.__CRi].index;
			}
			if (ix > 0)
			{
				total.unshift(total.splice(ix, 1)[0]);
			}
		}
		if (total.length == 0)
		{
			total.push(this.tabsToString(this.mLastWindowClosed.tabs));
			activeTab = this.mLastWindowClosed.index;
		}
		total.unshift(this.mCookies.join("\n"));
		
		var header = activeTab + " ~~ Crash Recovery ~~~~" + ((this.mLoadState > 0)?" " + (new Date()).toLocaleString():"") + ((this.mCookies.length)?"\n":"");
		
		return header + total.join(this.mTabSeparator);
	},

/* ........ Restoring Functionality .............. */

	restoreWindow: function(aWindow, aState, aOverwriteTabs)
	{
		var browser = aWindow.getBrowser();
		var tabs = aState.split(this.mTabSeparator);
		this.mCookies = tabs.shift().split("\n");
		var activeTab = parseInt(this.mCookies.shift());
		var currentTabs = [];
		var openTabCount = browser.browsers.length;
		
		this.restoreCookies(this.mCookies);
		
		for (var t = 0; t < tabs.length; t++)
		{
			currentTabs.push((t < openTabCount && aOverwriteTabs)?browser.mTabContainer.childNodes[t]:browser.addTab());
			if (!this.mLoadState && browser.browsers.length > 1)
			{
				browser.moveTabTo(currentTabs[currentTabs.length - 1], t);
			}
		}
		for (var i = currentTabs.length; i < openTabCount && aOverwriteTabs; i++)
		{
			browser.removeTab(i);
		}
		
		aWindow.setTimeout(function(aCRS) { aCRS.restoreHistory(aWindow, tabs, currentTabs, (aOverwriteTabs)?activeTab + 1:0, aOverwriteTabs); }, 100, this);
	},

	restoreHistory: function(aWindow, aTabs, aCurrentTabs, aSelectTab, aCount)
	{
		var postdata_limit = this.getPref(".postdata", -1);
		
		for (var t = 0; t < aTabs.length; t++)
		{
			var browser = aWindow.getBrowser().getBrowserForTab(aCurrentTabs[t]);
			var webNav = browser.webNavigation;
			
			try
			{
				var hist = webNav.sessionHistory;
				if (!hist) throw new Error();
			}
			catch (ex)
			{
				aCount = (aCount || 0) + 1;
				if (aCount < 10)
				{
					aWindow.setTimeout(function(aCRS) { aCRS.restoreHistory(aWindow, aTabs, aCurrentTabs, aSelectTab, aCount); }, 100, this);
				}
				return;
			}
			
			hist = webNav.sessionHistory;
			hist.QueryInterface(Components.interfaces.nsISHistoryInternal);
			if (hist.count > 0)
			{
				hist.PurgeHistory(hist.count);
			}
			
			var tabData = aTabs[t].split("\n");
			if (!/^(\d+) zoom:(\d+(?:\.\d+)?)/.test(tabData.shift()))
			{
				continue;
			}
			var activeIndex = parseInt(RegExp.$1);
			var textZoom = parseFloat(RegExp.$2);
			var activeScrollPosition = [0, 0];
			
			var tabFrameset = (tabData.length > 0 && /^@frameset /.test(tabData[tabData.length - 1]))?tabData.pop().substr(10):null;
			var tabText = (tabData.length > 0 && /^@text /.test(tabData[tabData.length - 1]))?tabData.pop().substr(6):null;
			
			for (var i = 0; i < tabData.length; i++)
			{
				var entries = tabData[i].split(" ");
				var entry = Components.classes["@mozilla.org/browser/session-history-entry;1"].createInstance(Components.interfaces.nsISHEntry);
				entry.setURI(this.mIOService.newURI(entries[0], null, null));
				entry.loadType = Components.interfaces.nsIDocShellLoadInfo.loadHistory;
				
				var scrollPosition = entries[1].split(",");
				scrollPosition = [parseInt(scrollPosition[0]), parseInt(scrollPosition[1])];
				entry.setScrollPosition(scrollPosition[0], scrollPosition[1]);
				
				entries[2] = decodeURI(entries[2] || "");
				if (entries[2] && (entries[2].length < postdata_limit || postdata_limit == -1))
				{
					var stream = Components.classes["@mozilla.org/io/string-input-stream;1"].createInstance(Components.interfaces.nsIStringInputStream);
					stream.setData(decodeURI(entries[2]), -1);
					entry.postData = stream;
				}
				hist.addEntry(entry, true);
				
				if (i == activeIndex)
				{
					activeScrollPosition = scrollPosition;
				}
			}
			
			if (t == aSelectTab - 1)
			{
				aWindow.getBrowser().selectedTab = aCurrentTabs[t];
			}
			try
			{
				webNav.gotoIndex(activeIndex);
			}
			catch (ex) { }
			
			browser.markupDocumentViewer.textZoom = textZoom;
			
			if (activeScrollPosition[0] || activeScrollPosition[1] || tabText || tabFrameset)
			{
				var id = "__CR" + t + "_" + (new Date()).valueOf();
				aWindow[id] = eval(this.restoreDocument.toString().replace(/__ID__/g, id).replace(/__SCROLL_POSITION__/g, activeScrollPosition.join(", ")).replace(/__TEXT_ARRAY__/g, (tabText)?"\"" + tabText + "\".split(' ')":"[]").replace(/__FRAMESET_ARRAY__/g, (tabFrameset)?"\"" + tabFrameset + "\".split(' ')":"[]"));
				aWindow[id + "_target"] = browser;
				browser.addEventListener("load", aWindow[id], true);
			}
		}
		
		this.onLoad_done(aWindow);
	},

	restoreDocument: function(aEvent)
	{
		function restoreTextData(aBrowser)
		{
			__TEXT_ARRAY__.forEach(function(aEntry) {
				if (/^(#?)([\w-]+)=(.*)$/.test(aEntry))
				{
					var document = aBrowser.document;
					var node = (RegExp.$1)?document.getElementById(RegExp.$2):document.getElementsByName(RegExp.$2)[0] || null;
					if (node && "value" in node)
					{
						aBrowser.setTimeout(function(aNode, aValue) {
							aNode.value = aValue;
						}, 0, node, decodeURI(RegExp.$3));
						var document = browser.top.document;
						if (!document.__CR_data)
						{
							document.__CR_data = {};
						}
						document.__CR_data[RegExp.$1 + RegExp.$2] = decodeURI(RegExp.$3);
					}
				}
			});
		}
		
		if (aEvent && aEvent.originalTarget && aEvent.originalTarget.defaultView == aEvent.originalTarget.defaultView.top)
		{
			var browser = aEvent.originalTarget.defaultView;
			browser.scrollTo(__SCROLL_POSITION__);
			restoreTextData(browser);
			
			__FRAMESET_ARRAY__.forEach(function(aEntry, aIx) {
				var id = "__CRf" + aIx + "_" + (new Date()).valueOf();
				browser[id] = function() {
					restoreTextData(browser.frames[aIx]);
					
					browser[id + "_target"].removeEventListener("load", browser[id], true);
					delete browser[id + "_target"];
					delete browser[id];
				};
				browser[id + "_target"] = browser.frames[aIx].frameElement;
				browser.frames[aIx].frameElement.addEventListener("load", browser[id], true);
				browser.frames[aIx].location.href = aEntry;
			});
			
			aWindow.__ID___target.removeEventListener("load", aWindow.__ID__, true);
			delete aWindow.__ID___target;
			delete aWindow.__ID__;
		}
	},

	restoreCookies: function(aCookies)
	{
		var cookieService = Components.classes["@mozilla.org/cookieService;1"].getService(Components.interfaces.nsICookieService);
		
		for (var i = 0; i < aCookies.length; i++)
		{
			var cookie = aCookies[i].split(" ");
			if (cookie.length == 2)
			{
				cookieService.setCookieString(this.mIOService.newURI(cookie[0], null, null), null, decodeURI(cookie[1]) + "expires=0;", null);
			}
		}
	},

	// code adapted from Danil Ivanov's "Cache Fixer" extension
	restoreCache: function()
	{
		var cache = Components.classes["@mozilla.org/file/directory_service;1"].getService(Components.interfaces.nsIProperties).get("ProfLD", Components.interfaces.nsILocalFile);
		cache.append("Cache");
		cache.append("_CACHE_MAP_");
		if (!cache.exists())
		{
			return;
		}
		
		var uri = this.mIOService.getProtocolHandler("file").QueryInterface(Components.interfaces.nsIFileProtocolHandler).newFileURI(cache);
		var channel = this.mIOService.newChannelFromURI(uri);
		var input = Components.classes["@mozilla.org/binaryinputstream;1"].createInstance(Components.interfaces.nsIBinaryInputStream);
		
		input.setInputStream(channel.open());
		var result = input.readByteArray(input.available());
		input.close();
		try
		{
			channel.close();
		}
		catch (ex) { }
		
		if (result[15] != 1)
		{
			return;
		}
		result[15] = 0;
		
		var fout = Components.classes["@mozilla.org/network/file-output-stream;1"].getService(Components.interfaces.nsIFileOutputStream);
		fout.init(cache, 0x02 | 0x08 | 0x20, 0644, 0); // PR_WRONLY | PR_CREATE_FILE | PR_TRUNCATE
		var output = Components.classes["@mozilla.org/binaryoutputstream;1"].createInstance(Components.interfaces.nsIBinaryOutputStream);
		output.setOutputStream(fout);
		output.writeByteArray(result, result.length);
		output.flush();
		fout.close();
		output.close();
	},

	retryDownloads: function()
	{
		function LinkCheckerObject(aUrl, aFilename)
		{
			this.url = aUrl;
			this.filename = aFilename;
			
			this.onStartRequest = function(aRequest, aContext) { };
			this.onStopRequest = function(aRequest, aContext, aStatus)
			{
				if (aStatus == Components.results.NS_OK)
				{
					var file = Components.classes["@mozilla.org/file/local;1"].createInstance(Components.interfaces.nsILocalFile);
					file.initWithPath(this.filename);
					if (file.exists())
					{
						file.remove(false);
					}
					
					window.saveURL(this.url, this.filename, null, true, true);
				}
			};
		}
		
		var downloadManager = Components.classes["@mozilla.org/download-manager;1"].getService(Components.interfaces.nsIDownloadManager);
		var rdfService = Components.classes["@mozilla.org/rdf/rdf-service;1"].getService(Components.interfaces.nsIRDFService);
		
		var rdfContainer = Components.classes["@mozilla.org/rdf/container;1"].createInstance(Components.interfaces.nsIRDFContainer);
		var datasource = downloadManager.datasource;
		
		try
		{
			rdfContainer.Init(datasource, rdfService.GetResource("NC:DownloadsRoot"));
		}
		catch (ex) // SeaMonkey doesn't support saveURL without user interaction, so we stop here (it doesn't support downloadManager.datasource, either)
		{
			return;
		}
		
		var downloads = rdfContainer.GetElements();
		
		while (downloads.hasMoreElements())
		{
			var download = downloads.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
			
			var node = datasource.GetTarget(rdfService.GetResource(download.Value), rdfService.GetResource("http://home.netscape.com/NC-rdf#DownloadState"), true);
			if (node)
			{
				node.QueryInterface(Components.interfaces.nsIRDFInt);
			}
			
			if (!node || node.Value != Components.interfaces.nsIDownloadManager.DOWNLOAD_DOWNLOADING)
			{
				continue;
			}
			
			node = datasource.GetTarget(rdfService.GetResource(download.Value), rdfService.GetResource("http://home.netscape.com/NC-rdf#URL"), true);
			node.QueryInterface(Components.interfaces.nsIRDFResource);
			
			var url = node.Value;
			
			node = datasource.GetTarget(rdfService.GetResource(download.Value), rdfService.GetResource("http://home.netscape.com/NC-rdf#File"), true);
			node.QueryInterface(Components.interfaces.nsIRDFResource);
			
			var linkChecker = Components.classes["@mozilla.org/network/urichecker;1"].createInstance(Components.interfaces.nsIURIChecker);
			
			linkChecker.init(this.mIOService.newURI(url, null, null));
			linkChecker.loadFlags = Components.interfaces.nsIRequest.LOAD_BACKGROUND;
			linkChecker.asyncCheck(new LinkCheckerObject(url, node.Value), null);
		}
	},

/* ........ Disk Access .............. */

	saveStateDelayed: function(aWindow, aDelay)
	{
		aWindow.setTimeout(function(aCRS) {
			aCRS.saveState();
		}, aDelay || 2000, this);
	},

	saveState: function(aForce)
	{
		if (!aForce)
		{
			var delay = (new Date()).valueOf() - this.mLastSaveTime;
			
			if (delay > 2 * this.mSaveInterval)
			{
				this.mSavePending = false;
			}
			else if (this.mSavePending)
			{
				return;
			}
			else if (delay < this.mSaveInterval)
			{
				var window = this.mWindowMediator.getMostRecentWindow("navigator:browser");
				if (window)
				{
					this.mSavePending = true;
					window.setTimeout(function(aCRS) { aCRS.mSavePending = false; aCRS.saveState(); }, this.mSaveInterval - delay, this);
					return;
				}
			}
		}
		this.stateToDisk(this.getCurrentState());
		this.mLastSaveTime = (new Date()).valueOf();
	},

	stateToDisk: function(aState, aBackup)
	{
		var file = this.mDirectoryService.get("ProfD", Components.interfaces.nsILocalFile);
		
		file.append("crashrecovery." + ((aBackup)?"bak":"dat"));
		try
		{
			this.delFile(file);
			this.writeFile(file, aState);
		}
		catch (ex) { }
	},

	stateFromDisk: function()
	{
		var file = this.mDirectoryService.get("ProfD", Components.interfaces.nsILocalFile);
		
		file.append("crashrecovery.dat");
		
		return this.readFile(file) || "";
	},

	clearDisk: function()
	{
		var file = this.mDirectoryService.get("ProfD", Components.interfaces.nsILocalFile);
		file.append("crashrecovery.dat");
		this.delFile(file);
		file = file.parent;
		file.append("crashrecovery.bak");
		this.delFile(file);
	},

/* ........ Auxiliary Functions .............. */

	getNextIndex: function()
	{
		for (var count = 1; count in this.mWindows; count++);
		
		return count;
	},

	getWindows: function(aType)
	{
		var windows = [];
		var enu = this.mWindowMediator.getEnumerator(aType || null);
		
		while (enu.hasMoreElements())
		{
			windows.push(enu.getNext());
		}
		
		return windows;
	},

	forEachDocument: function(aFunction)
	{
		this.getWindows("navigator:browser").forEach(function(aWindow) {
			var browser = aWindow.getBrowser();
			for (var i = 0; i < browser.browsers.length; i++)
			{
				try
				{
					aFunction(aWindow, browser.browsers[i].contentDocument);
				}
				catch (ex) { }
			}
		});
	},

	tabsToString: function(aTabs)
	{
		return aTabs.map(function(aTabData) {
			return aTabData.join("\n") + ((aTabData.text)?"\n@text " + aTabData.text:"") + ((aTabData.frameset)?"\n@frameset " + aTabData.frameset:"");
		}).join(this.mTabSeparator);
	},

	doResumeSession: function()
	{
		return this.getPref("browser.startup.page", 1) == 2 || /^about:crashrecovery\b/.test(this.getPref("browser.startup.homepage", ""));
	},

	doRecoverSession: function()
	{
		if (!this.getPref(".restore_prompt", true))
		{
			return true;
		}
		
		var recover = true;
		var dontPrompt = { value: false };
		try
		{
			recover = this.mPromptService.confirmCheck(null, "Crash Recovery", this.mStringBundle.GetStringFromName("restore_prompt"), this.mStringBundle.GetStringFromName("restore_prompt_again"), dontPrompt);
		}
		catch (ex) { } // ignore missing stringbundles
		
		if (dontPrompt.value)
		{
			this.setPref(".restore_prompt", false);
		}
		
		return recover;
	},

	isCmdLineEmpty: function(aWindow)
	{
		var homepage = this.getPref("browser.startup.homepage", "") + "|about:blank";
		try
		{
			homepage += "|" + Components.classes["@mozilla.org/browser/global-history;1"].getService(Components.interfaces.nsIBrowserHistory).lastPageVisited;
		}
		catch (ex) { }
		var homepageOnly = true;
		
		if (aWindow.arguments)
		{
			for (var i = 0; i < aWindow.arguments.length; i++)
			{
				if (homepage.indexOf(aWindow.arguments[i]) == -1)
				{
					homepageOnly = false;
				}
				else
				{
					aWindow.arguments.splice(i--, 1);
				}
			}
		}
		
		return homepageOnly;
	},

/* ........ Storage API .............. */

	getPref: function(aName, aDefault)
	{
		if (/^\./.test(aName))
		{
			aName = "extensions.crashrecovery" + aName;
		}
		
		var pb = this.mPrefBranch;
		switch (pb.getPrefType(aName))
		{
		case pb.PREF_STRING:
			return pb.getCharPref(aName);
		case pb.PREF_BOOL:
			return pb.getBoolPref(aName);
		case pb.PREF_INT:
			return pb.getIntPref(aName);
		default:
			return aDefault;
		}
	},

	setPref: function(aName, aValue)
	{
		if (/^\./.test(aName))
		{
			aName = "extensions.crashrecovery" + aName;
		}
		
		switch (typeof(aValue))
		{
		case "boolean":
			this.mPrefBranch.setBoolPref(aName, aValue);
			break;
		case "number":
			this.mPrefBranch.setIntPref(aName, parseInt(aValue));
			break;
		default:
			this.mPrefBranch.setCharPref(aName, "" + aValue);
			break;
		}
	},

	delPref: function(aName)
	{
		if (/^\./.test(aName))
		{
			aName = "extensions.crashrecovery" + aName;
		}
		
		this.mPrefBranch.deleteBranch(aName);
	},

	readFile: function(aFile)
	{
		try
		{
			var stream = Components.classes["@mozilla.org/network/file-input-stream;1"].createInstance(Components.interfaces.nsIFileInputStream);
			stream.init(aFile, 0x01, 0, 0);
			var sstream = Components.classes["@mozilla.org/scriptableinputstream;1"].createInstance(Components.interfaces.nsIScriptableInputStream);
			sstream.init(stream);
			
			var data = sstream.read(-1).replace(/\r?\n/g, "\n");
			sstream.close();
			stream.close();
			
			return this.mUnicodeConverter.ConvertToUnicode(data);
		}
		catch (ex) { }
		
		return null;
	},

	writeFile: function(aFile, aData)
	{
		var stream = Components.classes["@mozilla.org/network/file-output-stream;1"].createInstance(Components.interfaces.nsIFileOutputStream);
		aData = this.mUnicodeConverter.ConvertFromUnicode(aData).replace(/\n/g, "\r\n");
		
		aFile.create(Components.interfaces.nsIFile.NORMAL_FILE_TYPE, 0600);
		stream.init(aFile, 0x02 | 0x08 | 0x20, 0600, 0);
		stream.write(aData, aData.length);
		stream.flush();
		stream.close();
	},

	delFile: function(aFile)
	{
		try
		{
			aFile.remove(false);
		}
		catch (ex) { }
	},

/* ........ QueryInterface .............. */

	QueryInterface: function(aIID)
	{
		if (!aIID.equals(Components.interfaces.nsISupports) && !aIID.equals(Components.interfaces.nsIObserver) && !aIID.equals(this.mCID))
		{
			throw Components.results.NS_ERROR_NO_INTERFACE;
		}
		
		return this;
	}
};

/* :::::::: Miscellanea ::::::::::::::: */

function getStringBundle(aURI)
{
	return Components.classes["@mozilla.org/intl/stringbundle;1"].getService(Components.interfaces.nsIStringBundleService).createBundle(aURI, Components.classes["@mozilla.org/intl/nslocaleservice;1"].getService(Components.interfaces.nsILocaleService).getApplicationLocale());
}

function debug(aMsg)
{
	Components.classes["@mozilla.org/consoleservice;1"].getService(Components.interfaces.nsIConsoleService).logStringMessage(("" + aMsg || " ").replace(/\S{80}/g, "$&\n"));
}

/* :::::::: about:crashrecovery ::::::::::::::: */

var // const
CrashRecoveryAbout = {
	mCID: Components.ID("{1280606b-2510-4fe0-97ef-9b5a22eafe6c}"),
	mContractID: "@mozilla.org/network/protocol/about;1?what=crashrecovery",
	mClassName: "crashrecoveryAbout",

	mIOService: Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService),

	newChannel: function(aURI)
	{
		var uri = aURI.spec.substr(20);
		
		if (/^(?!about:(?!blank$)).+/.test(uri))
		{
			uri = "javascript:location.href=\"" + uri.replace(/"/g, "%22") + "\";";
		}
		else if (!uri)
		{
			uri = "data:," + encodeURI(CrashRecoveryService.stateFromDisk());
		}
		
		return this.mIOService.newChannel(uri, null, aURI);
	},

	QueryInterface: function(aIID)
	{
		if (!aIID.equals(Components.interfaces.nsISupports) && !aIID.equals(Components.interfaces.nsIAboutModule))
		{
			throw Components.results.NS_ERROR_NO_INTERFACE;
		}
		
		return this;
	}
}

/* :::::::: Service Registration & Initialization ::::::::::::::: */

function ObjectFactory(aObj)
{
	this.mObj = aObj;
	this.mCID = aObj.mCID;
	this.mContractID = aObj.mContractID;
	this.mClassName = aObj.mClassName;
	
	this.createInstance = function(aOuter, aIID) {
		if (aOuter != null)
		{
			throw Components.results.NS_ERROR_NO_AGGREGATION;
		}
		
		return this.mObj.QueryInterface(aIID);
	};
	this.lockFactory = function(aLock) { };
	this.QueryInterface = function(aIID) {
		if (!aIID.equals(Components.interfaces.nsISupports) && !aIID.equals(Components.interfaces.nsIFactory))
		{
			throw Components.results.NS_ERROR_NO_INTERFACE;
		}
		
		return this;
	};
}

var // const
CrashRecoveryModule = {
	mFactories: [new ObjectFactory(CrashRecoveryService), new ObjectFactory(CrashRecoveryAbout)],

	getClassObject: function(aCompMgr, aCID, aIID)
	{
		for (var ix in this.mFactories)
		{
			if (aCID.equals(this.mFactories[ix].mCID))
			{
				return this.mFactories[ix].QueryInterface(aIID);
			}
		}
		
		throw Components.results.NS_ERROR_NOT_REGISTERED;
	},

	registerSelf: function(aCompMgr, aFileSpec, aLocation, aType)
	{
		aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
		for (var ix in this.mFactories)
		{
			aCompMgr.registerFactoryLocation(this.mFactories[ix].mCID, this.mFactories[ix].mClassName, this.mFactories[ix].mContractID, aFileSpec, aLocation, aType);
		}
		
		var catMan = Components.classes["@mozilla.org/categorymanager;1"].getService(Components.interfaces.nsICategoryManager);
		catMan.addCategoryEntry("app-startup", CrashRecoveryService.mClassName, "service," + CrashRecoveryService.mContractID, true, true, null);
	},

	unregisterSelf: function(aCompMgr, aLocation, aType)
	{
		aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
		for (var ix in this.mFactories)
		{
			aCompMgr.unregisterFactoryLocation(this.mFactories[ix].mCID, aLocation);
		}
		
		var catMan = Components.classes["@mozilla.org/categorymanager;1"].getService(Components.interfaces.nsICategoryManager);
		catMan.deleteCategoryEntry("app-startup", "service," + CrashRecoveryService.mContractID, true);
	},

	canUnload: function(aCompMgr)
	{
		return true;
	},

	QueryInterface: function(aIID)
	{
		if (!aIID.equals(Components.interfaces.nsISupports) && !aIID.equals(Components.interfaces.nsIModule))
		{
			throw Components.results.NS_ERROR_NO_INTERFACE;
		}
		
		return this;
	}
};

function NSGetModule(aComMgr, aFileSpec)
{
	return CrashRecoveryModule;
}
